/* ----------------------------------------------------------------------------
sv.cc

Generates a random sample or plot points from a scale mixture of a normal
by a log normal.

Copyright (C) 1993, 1994, 1997, 2002, 2006.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659 
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "libscl.h"
using namespace scl;
using namespace std;

#define POINTS 15

const static REAL weight[POINTS] = {
  .000000001522476,
  .000001059115548,
  .000100004441232,
  .002778068842913,
  .030780033872546,
  .158488915795936,
  .412028687498898,
  .564100308726418,
  .412028687498898,
  .158488915795936,
  .030780033872546,
  .002778068842913,
  .000100004441233,
  .000001059115548,
  .000000001522476
};

const static REAL abcissa[POINTS] = {
  -4.499990707309390,
  -3.669950373404451,
  -2.967166927905604,
  -2.325732486173856,
  -1.719992575186489,
  -1.136115585210921,
   -.565069583255576,
    .000000000000000,
    .565069583255576,
   1.136115585210921,
   1.719992575186488,
   2.325732486173858,
   2.967166927905603,
   3.669950373404451,
   4.499990707309388
};

char       *progname;

void       usage_msg(void);
void       process_file(istream& cfile);
void       badflag(char * text);

void badflag(char * text) {
  cerr << "Error, "<< progname <<", unknown flag or value missing: "
       << text << ".\n";
  exit(1);
}

REAL       normal(REAL z, int isw);

INTEGER    sample_size;         // Set with flag -n
int        plot = 0;            // The flag -p sets to true
int        eformat = 0;         // The flag -e sets to true
char*      parm = 0;            // Set with flag -w, defaults to none.
REAL       xmin = -3.0;         // Set with flag -xmin, defaults otherwise.
REAL       xmax =  3.0;         // Set with flag -xmax, defaults otherwise.
INT_32BIT  seed = 349857;       // Can be changed with flag -s

void usage_msg()
{
  cerr << '\n'
    <<"usage: " << progname << " -n value -r1 value -r2 value -r3 value [options]\n\n"
    <<"flags:   -n  value      number of observations or number of plot points\n"
    <<"                        the flag -m is a synonym\n"
    <<"         -r1 value      the value of rho[1], mean of mixed density\n" 
    <<"         -r2 value      the value of rho[2], mean of mixing density\n" 
    <<"         -r3 value      the value of rho[3], std dev of mixing density\n\n" 
    <<"options: -s  value      supply an integer seed rather than accept default\n"
    <<"         -p             generate plot points\n"
    <<"         -e             output observations in e format with full precision\n"
    <<"         -w  filename   write r1, r2, r3 to filename\n"
    <<"         -xmin  value   minimum plot point\n"
    <<"         -xmax  value   maximum plot point\n\n"
    <<"examples:  "<<progname<<" -n 200 -r1 0.0 -r2 1.0 -r3 3.0  > out.dat\n"
    <<"           The output file contains y values, one per line.\n"
    <<"           "<<progname<<" -n 200 -r1 0.0 -r2 1.0 -r3 3.0 -p > out.dat\n"
    <<"           The output file contains y x pairs, one pair per line, y first.\n"
    <<"reference: Gallant, A. R., and G. Tauchen (1999), The Relative Efficiency\n"
    <<"           of Method of Moments Estimators, J. of Econometrics, forthcoming.\n"
    <<"comment:   The moments of the distribution determine rho: r1=mean,\n"
    <<"           r2=log(sigma)-r3*r3, and r3=sqrt(log(raw_kurtosis/3)/4).\n";
  exit(1);
}

// Template for a program that reads flags and files from the command line and 
// processes those files sequentially.  Multiple flags, grouping of flags, and 
// conditional read of next file after a flag are supported.  

int main(int argc, char *argp[], char *envp[])
{
  int      argx, flagx, filex;
  char     flag;

  sample_size = 0;

  flag = envp[0][0];  // This supresses a compiler warning.

  progname = argp[0];

  REAL r1=REAL_MAX, r2=REAL_MAX, r3=REAL_MIN;

  // begin: loop over flags and input files; from = file1, file2, ... 
  // If there are no files, loop once.

  argx = 0;  filex = 1;  

  while (argx < argc) {

    if (argx==0) argx++;

    // begin: read flags, if any.

    while (argx < argc && (argp[argx])[0]=='-') {

      // Read flags, if any.  Note that flags can occur singly as -a -b 
      // or they can be grouped as -ab.  Just insert code within the
      // appropriate case and don't worry about it.

      for (flagx=1; (flag=(argp[argx])[flagx])!=0; flagx++) {

        // If a flag can have more than one letter, deal with it within the 
        // appropriate case by stripping off letters and incrementing flagx.

        // If a flag indicates that argp[argx+1] is a file that should be read 
        // then, within the appropriate case, open arg[argx+1] or otherwise 
        // deal with the filename arg[argx+1], assign argp[argx] to 
        // argp[argx+1], and increment argx.  

        // If a flag indicates that argp[argx+1] is a string containing the  
        // value of a parameter, etc., then, within the appropriate case, deal 
        // with the string arg[argx+1], assign argp[argx] to argp[argx+1], and 
        // increment argx.  

        switch (flag) {

        case 'p' :
          plot = 1;
          break;

        case 'e' :
          eformat = 1;
          break;

        case 'm' :
        case 'n' :
          sample_size = (INTEGER) atoi(argp[argx+1]); 
          argp[argx+1] = argp[argx];
          if (sample_size <= 0 ) {
            cerr << "sample_size must be positive\n"; 
	    badflag(&(argp[argx][flagx]));
          }
          argx++;
          break;

        case 'r' :
	  if (memcmp(&(argp[argx][flagx]),"r1",2)==0) {
	    r1 = (REAL) atof(argp[argx+1]);
          }
	  else if (memcmp(&(argp[argx][flagx]),"r2",2)==0) {
	    r2 = (REAL) atof(argp[argx+1]);
          }
	  else if (memcmp(&(argp[argx][flagx]),"r3",2)==0) {
	    r3 = (REAL) atof(argp[argx+1]);
          }
	  else {
	    badflag(&(argp[argx][flagx]));
          }
	  flagx++;
	  argp[argx+1] = argp[argx];
	  argx++;
	  break;
	  
        case 's' :
          seed = (INTEGER) atoi(argp[argx+1]); 
          argp[argx+1] = argp[argx];
          argx++;
          break;

        case 'w':
	  parm = argp[argx+1];
	  argp[argx+1] = argp[argx];
	  argx++;
	  break;

        case 'x' : 
          if (memcmp(&(argp[argx][flagx]),"xmin",4)==0) {
            flagx += 3;
            xmin = atof(argp[argx+1]); 
            argp[argx+1] = argp[argx];
          }
          else if (memcmp(&(argp[argx][flagx]),"xmax",4)==0) {
            flagx += 3;
            xmax = atof(argp[argx+1]); 
            argp[argx+1] = argp[argx];
          }
	  else {
            badflag(&(argp[argx][flagx]));
          }
          argx++;
          break;

        default :
          cerr << "Flag was ";
          while ((flag=(argp[argx])[flagx++])!=0) 
            cerr << flag;
          cerr << "\nError: unknown flag or other error.\n";
          exit(1);
          break;

        } // end: switch

      } // end: for 

    argx++;

    } // end: while

    // end: read flags, if any.


    if (filex == 1 && argx >= argc) {

      // Put code to process cin here.  This would include the case where no  
      // files are to be read; i.e. flag processing is all that is required.

      // Example of cin processing.
      // process_file(cin);

      if ( (sample_size == 0)||(r1==REAL_MAX)||(r2==REAL_MAX)||(r3==REAL_MIN) ) usage_msg();

      char str[80];

      if (parm) {
        ofstream parmfile(parm);
        if (!parmfile) {cerr << "Error, Cannot open " << parm << '\n'; exit(1);}
        if (eformat) {
          sprintf(str,"%30.18e\n",r1); parmfile << str;
	  sprintf(str,"%30.18e\n",r2); parmfile << str;
	  sprintf(str,"%30.18e\n",r3); parmfile << str;
        }
        else {
	  parmfile << r1 <<'\n'<< r2 <<'\n'<< r3 << '\n';
        }
        parmfile.close();
      }

      if (plot) 
        cerr << progname << ": " << sample_size << " plot points, "
	     << " r1 = " << r1 << ", r2 = " << r2 << ", r3 = " << r3 << '\n';
      else
        cerr << progname << ": " <<sample_size<<" observations, "
	     << " r1 = " << r1 << ", r2 = " << r2 << ", r3 = " << r3 << '\n';

      const REAL zero = 0.0;
      const REAL one = 1.0;
      const REAL root_two = 1.414213562373095;
      const REAL root_pi  = 1.772453850905516;

      INTEGER i,j,k;

      REAL    mean;
      REAL    std_err[POINTS];
      REAL    prob[POINTS];

      mean = r1;

      for (j=0; j<POINTS; j++) {
	prob[j] = weight[j]/root_pi;
        std_err[j] = (REAL) exp( (double) (root_two * abcissa[j] * r3 + r2) );
      }

      if (plot) {
        REAL inc = (xmax-xmin)/(sample_size-1.0);
        REAL x, y, z;
        for (j=0; j<sample_size; j++) {
          x = xmin + inc*j;
          y = 0.0;
          for (i=0; i<POINTS; i++) {
            z = (x - mean)/std_err[i];
            y += prob[i]*normal(z,1)/std_err[i];
          }
          cout << y << "  " << x << '\n';
        }
      }
      else {
        INT_32BIT * ix;
        ix = &seed;

        REAL low, high;
        REAL x, y, z;

        for (k=0; k<sample_size; k++) {
          x = ran(ix);
          for (low=0.0, high=0.0, i=0; i<POINTS; i++) { 
            high += prob[i];
            if ( (low<=x) && (x<=high) ) { j=i; break; }
            low += prob[i];
          }

          z = unsk(ix);
          y = mean + std_err[j]*z;

          if (eformat) {
            sprintf(str,"%30.18e\n",y);
            cout << str;
          }
          else {
            cout << y << '\n';
          }
        }
      }
      exit(0);
    }

    usage_msg();

    // begin: open input file, if any.

    ifstream infile(argp[argx]);
    if (!infile) {
      cerr << "Error, " << argp[0] << ", Cannot open " << argp[argx] << "\n";
      exit(1);
    } 
    // istream& from = infile;

    // end: open input file, if any.

    // Put code to process a file here.

    // Example of file processing
    // process_file(from);

    infile.close();
    filex++;
    argx++;
  } 

  // end: loop over flags and input files.

  exit(0);

}

void process_file(istream& from)
{ char ch; while (from.get(ch)) cout << ch; }

REAL normal(REAL z, int isw)
// Returns the value of the normal kernel at z if isw=1 and its 
// convolution with itself if isw=2.
{
  REAL const two = 2.0;
  REAL const root_two_pi = 2.5066282746310005024;
  REAL const root_two    = 1.4142135623730950488;
  REAL       h;
  REAL       tmp;

  switch (isw) {

    case 1 :
     tmp = -z*z/two;
     h = (REAL) exp( (double) tmp );
     h /= root_two_pi;
     return h;

    case 2 :
     tmp = -z*z/two;
     tmp /= two;
     h = (REAL) exp( (double) tmp );
     h /= root_two_pi;
     h /= root_two;
     return h;

    default :
      cerr << "\nError: bad data to REAL normal(REAL z, int isw)\n";
      exit(1);
   }

   // The next two statemants can't be reached but stop a compiler warning.
   cerr << "\nError: bad data to REAL normal(REAL z, int isw)\n";
   return 1;  
}
